"""Utilities for algorithms retrieval."""

from typing import Dict, List

from ..algorithms.registry import ApplicationsRegistry, ConfigurationTuple
from ..configuration import reset_logging_root_logger

reset_logging_root_logger()

AVAILABLE_ALGORITHMS = sorted(
    ApplicationsRegistry.list_available(),
    key=lambda algorithm: (
        algorithm["domain"],
        algorithm["algorithm_type"],
        algorithm["algorithm_name"],
        algorithm["algorithm_application"],
        algorithm["algorithm_version"],
    ),
)
AVAILABLE_ALGORITHMS_CATEGORIES = {
    category: sorted(set([algorithm[category] for algorithm in AVAILABLE_ALGORITHMS]))
    for category in ["domain", "algorithm_type"]
}


def filter_algorithm_applications(
    algorithms: List[Dict[str, str]], filters: Dict[str, str]
) -> List[Dict[str, str]]:
    """
    Returning algorithms with given filters.

    Args:
        algorithms (List[Dict[str, str]]): a list of algorithm applications as dictionaries.
        filters (Dict[str, str]): the filters to apply.

    Returns:
        the filtered algorithms.
    """
    return [
        application
        for application in algorithms
        if all(
            [
                application[filter_type] == filter_value if filter_value else True
                for filter_type, filter_value in filters.items()
            ]
        )
    ]


def get_configuration_tuples(
    algorithms: List[Dict[str, str]]
) -> List[ConfigurationTuple]:
    """
    Returning configuration tuples from a list of applications.

    Args:
        algorithms (List[Dict[str, str]]): a list of algorithm applications as dictionaries.

    Returns:
        the configuration tuples.
    """
    return sorted(
        list(
            {
                ConfigurationTuple(
                    **{
                        key: value
                        for key, value in algorithm.items()
                        if key in ConfigurationTuple.__annotations__.keys()
                    }
                )
                for algorithm in algorithms
            }
        )
    )



"""Regression Transformer initialization."""
from .core import (
    RegressionTransformer,
    RegressionTransformerMolecules,
    RegressionTransformerProteins,
)

__all__ = [
    "RegressionTransformer",
    "RegressionTransformerMolecules",
    "RegressionTransformerProteins",
]



"""RegressionTransformer tests."""

import pickle
from typing import ClassVar, Type

import pytest

from gt4sd.algorithms.conditional_generation.regression_transformer import (
    RegressionTransformer,
    RegressionTransformerMolecules,
    RegressionTransformerProteins,
)
from gt4sd.algorithms.core import AlgorithmConfiguration
from gt4sd.algorithms.registry import ApplicationsRegistry
from gt4sd.tests.utils import GT4SDTestSettings

test_settings = GT4SDTestSettings.get_instance()


def get_classvar_type(class_var):
    """Extract type from ClassVar type annotation: `ClassVar[T]] -> T`."""
    return class_var.__args__[0]


@pytest.mark.parametrize(
    "config_class, algorithm_type, domain, algorithm_name",
    [
        (
            RegressionTransformerMolecules,
            "conditional_generation",
            "materials",
            RegressionTransformer.__name__,
        ),
        (
            RegressionTransformerProteins,
            "conditional_generation",
            "materials",
            RegressionTransformer.__name__,
        ),
    ],
)
def test_config_class(
    config_class: Type[AlgorithmConfiguration],
    algorithm_type: str,
    domain: str,
    algorithm_name: str,
):
    assert config_class.algorithm_type == algorithm_type
    assert config_class.domain == domain
    assert config_class.algorithm_name == algorithm_name

    for keyword, type_annotation in config_class.__annotations__.items():
        if keyword in ("algorithm_type", "domain", "algorithm_name"):
            assert type_annotation.__origin__ is ClassVar  # type: ignore
            assert str == get_classvar_type(type_annotation)


@pytest.mark.parametrize(
    "config_class",
    [
        (RegressionTransformerMolecules),
        (RegressionTransformerProteins),
    ],
)
def test_config_instance(config_class: Type[AlgorithmConfiguration]):
    config = config_class()  # type:ignore
    assert config.algorithm_application == config_class.__name__


@pytest.mark.parametrize(
    "config_class",
    [
        (RegressionTransformerMolecules),
        (RegressionTransformerProteins),
    ],
)
def test_available_versions(config_class: Type[AlgorithmConfiguration]):
    versions = config_class.list_versions()
    assert "v0" in versions


@pytest.mark.parametrize(
    "config, example_target, algorithm, params",
    [
        pytest.param(
            RegressionTransformerMolecules,
            "<esol>[MASK][MASK][MASK][MASK][MASK]|[Cl][C][Branch1_2][Branch1_2][=C][Branch1_1][C][Cl][Cl][Cl]",
            RegressionTransformer,
            {"search": "greedy", "num_samples": 1},
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerMolecules,
            "<esol>-3.499|[C][C][MASK][MASK][MASK][C][Br]",
            RegressionTransformer,
            {"search": "sample", "temperature": 2.0, "num_samples": 5},
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerProteins,
            "<stab>[MASK][MASK][MASK][MASK][MASK]|GSQEVNSNASPEEAEIARKAGATTWTEKGNKWEIRI",
            RegressionTransformer,
            {"search": "greedy", "num_samples": 1},
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerProteins,
            "<stab>1.1234|TTIKNG[MASK][MASK][MASK]YTVPLSPEQAAK[MASK][MASK][MASK]KKRWPDYEVQIHGNTVKVT",
            RegressionTransformer,
            {"search": "sample", "temperature": 2.0, "num_samples": 5},
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
    ],
)
def test_generation_via_import(config, example_target, algorithm, params):
    num_samples = params.pop("num_samples", 1)
    regression_transformer = algorithm(
        configuration=config(**params), target=example_target
    )
    items = list(regression_transformer.sample(num_samples))
    assert len(items) == num_samples


@pytest.mark.parametrize(
    "algorithm_application, target, params",
    [
        pytest.param(
            RegressionTransformerMolecules.__name__,
            "<esol>[MASK][MASK][MASK][MASK][MASK]|[Cl][C][Branch1_2][Branch1_2][=C][Branch1_1][C][Cl][Cl][Cl]",
            {"search": "greedy", "num_samples": 1},
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerMolecules.__name__,
            "<esol>-3.499|[C][C][MASK][MASK][MASK][C][Br]",
            {"search": "sample", "temperature": 2.0, "num_samples": 5},
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerProteins.__name__,
            "<stab>[MASK][MASK][MASK][MASK][MASK]|GSQEVNSNASPEEAEIARKAGATTWTEKGNKWEIRI",
            {"search": "greedy", "num_samples": 1},
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
        pytest.param(
            RegressionTransformerProteins.__name__,
            "<stab>1.1234|TTIKNG[MASK][MASK][MASK]YTVPLSPEQAAK[MASK][MASK][MASK]KKRWPDYEVQIHGNTVKVT",
            {"search": "sample", "temperature": 2.0, "num_samples": 5},
            marks=pytest.mark.skipif(test_settings.gt4sd_ci, reason="high_memory"),
        ),
    ],
)
def test_generation_via_registry(target, algorithm_application, params):
    num_samples = params.pop("num_samples", 1)
    regression_transformer = ApplicationsRegistry.get_application_instance(
        target=target,
        algorithm_type="conditional_generation",
        domain="materials",
        algorithm_name=RegressionTransformer.__name__,
        algorithm_application=algorithm_application,
        **params,
    )
    items = list(regression_transformer.sample(num_samples))
    assert len(items) == num_samples


@pytest.mark.parametrize(
    "config_class",
    [
        (RegressionTransformerMolecules),
        (RegressionTransformerProteins),
    ],
)
def test_configuration_pickable(config_class: Type[AlgorithmConfiguration]):
    # implementation
    obj = config_class(algorithm_version="test")

    # ---
    import inspect

    inspect.getmodule(config_class)
    # ---
    pickled_obj = pickle.dumps(obj)
    restored_obj = pickle.loads(pickled_obj)
    assert restored_obj.algorithm_version == "test"
    assert restored_obj == obj

    # registered
    Config = ApplicationsRegistry.get_application(
        algorithm_type="conditional_generation",
        domain="materials",
        algorithm_name=RegressionTransformer.__name__,
        algorithm_application=config_class.__name__,
    ).configuration_class

    obj = Config(algorithm_version="test")
    pickled_obj = pickle.dumps(obj)
    restored_obj = pickle.loads(pickled_obj)

    assert restored_obj.algorithm_version == "test"
    assert restored_obj == obj



"""Generic utils for pytorch."""

from typing import List, Optional, Union

import torch


def get_gpu_device_names() -> List[str]:
    """Get GPU device names as a list.

    Returns:
        names of available GPU devices.
    """
    gpu_device_names = []
    if torch.cuda.is_available():
        gpu_device_names = [
            "cuda:{index}" for index in range(torch.cuda.device_count())
        ]
    return gpu_device_names


def claim_device_name() -> str:
    """Claim a device name.

    Returns:
        device name, if on GPU is available returns CPU.
    """
    device_name = "cpu"
    gpu_device_names = get_gpu_device_names()
    if len(gpu_device_names) > 0:
        device_name = gpu_device_names[0]
    return device_name


def get_device() -> torch.device:
    """
    Get device dynamically.
    """
    return torch.device("cuda" if torch.cuda.is_available() else "cpu")


def device_claim(device: Optional[Union[torch.device, str]] = None) -> torch.device:
    """
    Satidfy a device claim.

    Args:
        device: device where the inference
            is running either as a dedicated class or a string. If not provided is inferred.

    Returns:
        torch.device: the claimed device or a default one.
    """
    if isinstance(device, str):
        device = torch.device(device)
    device = (
        get_device()
        if (device is None or not isinstance(device, torch.device))
        else device
    )
    return device


def get_device_from_tensor(tensor: torch.Tensor) -> torch.device:
    """Get the device from a tensor.

    Args:
        tensor: a tensor.

    Returns:
        the device.
    """
    device_id = tensor.get_device()
    device = "cpu" if device_id < 0 else f"cuda:{device_id}"
    return device_claim(device)



"""Module initialization for gt4sd."""
from ..extras import EXTRAS_ENABLED

# NOTE: here we import the applications to register them
from .conditional_generation.guacamol.core import (  # noqa: F401
    AaeGenerator,
    GraphGAGenerator,
    GraphMCTSGenerator,
    OrganGenerator,
    SMILESGAGenerator,
    SMILESLSTMHCGenerator,
    SMILESLSTMPPOGenerator,
    VaeGenerator,
)
from .conditional_generation.key_bert.core import KeyBERTGenerator  # noqa: F401
from .conditional_generation.paccmann_rl.core import (  # noqa: F401
    PaccMannRLOmicBasedGenerator,
    PaccMannRLProteinBasedGenerator,
)
from .conditional_generation.regression_transformer.core import (  # noqa: F401
    RegressionTransformerMolecules,
    RegressionTransformerProteins,
)
from .conditional_generation.reinvent.core import ReinventGenerator  # noqa: F401
from .conditional_generation.template.core import TemplateGenerator  # noqa: F401
from .controlled_sampling.advanced_manufacturing.core import (  # noqa: F401
    CatalystGenerator,
)
from .controlled_sampling.paccmann_gp.core import PaccMannGPGenerator  # noqa: F401
from .generation.hugging_face.core import (  # noqa: F401
    HuggingFaceCTRLGenerator,
    HuggingFaceGPT2Generator,
    HuggingFaceOpenAIGPTGenerator,
    HuggingFaceTransfoXLGenerator,
    HuggingFaceXLMGenerator,
    HuggingFaceXLNetGenerator,
)
from .generation.pgt.core import (  # noqa: F401
    PGTCoherenceChecker,
    PGTEditor,
    PGTGenerator,
)
from .generation.polymer_blocks.core import PolymerBlocksGenerator  # noqa: F401
from .generation.torchdrug.core import TorchDrugGCPN, TorchDrugGraphAF  # noqa: F401
from .prediction.paccmann.core import PaccMann  # noqa: F401
from .prediction.topics_zero_shot.core import TopicsPredictor  # noqa: F401

# extras requirements
if EXTRAS_ENABLED:
    from .controlled_sampling.class_controlled_sampling.core import (  # noqa: F401
        PAG,
        CogMol,
    )
    from .generation.molgx.core import MolGXQM9Generator  # noqa: F401



"""Argument parser for training pipelines."""


import dataclasses
import re
from argparse import ArgumentTypeError
from enum import Enum
from functools import partial
from typing import Any, List, NewType, Optional, Type, Union

from transformers import HfArgumentParser


def none_checker_bool(val: Union[bool, str]) -> Union[bool, None]:
    """Check given bool argument for None.

    Args:
        val: model arguments passed to the configuration.
    Returns:
        Bool value or None.
    """
    if not val:
        return None
    if isinstance(val, bool):
        return val
    if val.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif val.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise ArgumentTypeError(
            f"Truthy value expected: got {val} but expected one of yes/no, true/false, t/f, y/n, 1/0 (case insensitive)."
        )


def none_checker(val: Any, dtype: Type) -> Any:
    """Check given argument for None.

    Args:
        val: model arguments passed to the configuration.
        dtype: expected argument type.

    Returns:
        Value casted in the expected type or None.
    """
    if not val or val == "none":
        return None
    return dtype(val)


DataClass = NewType("DataClass", Any)  # type: ignore
DataClassType = NewType("DataClassType", Any)  # type: ignore


class ArgumentParser(HfArgumentParser):
    """ArgumentParser inherited from hf's parser with modified dataclass arguments addition for better handling of None values."""

    def _add_dataclass_arguments(self, dtype: DataClassType) -> None:
        """Add a dataclass arguments.

        Args:
            dtype: data class type.
        """

        if hasattr(dtype, "_argument_group_name"):
            parser = self.add_argument_group(dtype._argument_group_name)
        else:
            parser = self  # type: ignore
        for field in dataclasses.fields(dtype):
            if not field.init:
                continue
            field_name = f"--{field.name}"
            kwargs = field.metadata.copy()  # type: ignore
            # field.metadata is not used at all by Data Classes,
            # it is provided as a third-party extension mechanism.
            if isinstance(field.type, str):
                raise ImportError(
                    "This implementation is not compatible with Postponed Evaluation of Annotations (PEP 563),"
                    "which can be opted in from Python 3.7 with `from __future__ import annotations`."
                    "We will add compatibility when Python 3.9 is released."
                )
            typestring = str(field.type)
            for prim_type in (int, float, str):
                for collection in (List,):
                    if (
                        typestring == f"typing.Union[{collection[prim_type]}, NoneType]"  # type: ignore
                        or typestring == f"typing.Optional[{collection[prim_type]}]"  # type: ignore
                    ):
                        field.type = collection[prim_type]  # type: ignore
                if (
                    typestring == f"typing.Union[{prim_type.__name__}, NoneType]"
                    or typestring == f"typing.Optional[{prim_type.__name__}]"
                ):
                    field.type = prim_type

            if isinstance(field.type, type) and issubclass(field.type, Enum):
                kwargs["choices"] = [x.value for x in field.type]
                kwargs["type"] = type(kwargs["choices"][0])
                if field.default is not dataclasses.MISSING:
                    kwargs["default"] = field.default
                else:
                    kwargs["required"] = True
            elif field.type is bool or field.type == Optional[bool]:

                if field.default is True:
                    parser.add_argument(
                        f"--no_{field.name}",
                        action="store_false",
                        dest=field.name,
                        **kwargs,
                    )

                # Hack because type=bool in argparse does not behave as we want.
                kwargs["type"] = none_checker_bool
                if field.type is bool or (
                    field.default is not None
                    and field.default is not dataclasses.MISSING
                ):
                    # Default value is False if we have no default when of type bool.
                    default = (
                        False if field.default is dataclasses.MISSING else field.default
                    )
                    # This is the value that will get picked if we don't include --field_name in any way
                    kwargs["default"] = default
                    # This tells argparse we accept 0 or 1 value after --field_name
                    kwargs["nargs"] = "?"
                    # This is the value that will get picked if we do --field_name (without value)
                    kwargs["const"] = True
            elif (
                hasattr(field.type, "__origin__")
                and re.search(r"^typing\.List\[(.*)\]$", str(field.type)) is not None
            ):
                kwargs["nargs"] = "+"
                kwargs["type"] = partial(none_checker, dtype=field.type.__args__[0])
                assert all(
                    x == kwargs["type"] for x in field.type.__args__
                ), f"{field.name} cannot be a List of mixed types"
                if field.default_factory is not dataclasses.MISSING:  # type: ignore
                    kwargs["default"] = field.default_factory()  # type: ignore
                elif field.default is dataclasses.MISSING:
                    kwargs["required"] = True
            else:
                kwargs["type"] = partial(none_checker, dtype=field.type)
                if field.default is not dataclasses.MISSING:
                    kwargs["default"] = field.default
                elif field.default_factory is not dataclasses.MISSING:  # type: ignore
                    kwargs["default"] = field.default_factory()  # type: ignore
                else:
                    kwargs["required"] = True
            parser.add_argument(field_name, **kwargs)



"""Moses VAE training pipeline."""
import argparse
import logging
import os
from dataclasses import dataclass, field
from typing import Any, Dict

from guacamol_baselines.moses_baselines.vae_train import main

from ...core import TrainingPipelineArguments
from ..core import MosesTrainingArguments, MosesTrainingPipeline

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class MosesVAETrainingPipeline(MosesTrainingPipeline):
    """Moses VAE training pipelines."""

    def train(  # type: ignore
        self,
        training_args: Dict[str, Any],
        model_args: Dict[str, Any],
        dataset_args: Dict[str, Any],
    ) -> None:
        """Generic training function for Moses VAE training.

        Args:
            training_args: training arguments passed to the configuration.
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.
        """
        params = {**training_args, **model_args, **dataset_args}

        os.makedirs(os.path.dirname(params["model_save"]), exist_ok=True)
        os.makedirs(os.path.dirname(params["log_file"]), exist_ok=True)
        os.makedirs(os.path.dirname(params["config_save"]), exist_ok=True)
        os.makedirs(os.path.dirname(params["vocab_save"]), exist_ok=True)

        args = argparse.Namespace(**params)
        main(args)


@dataclass
class MosesVAEModelArguments(TrainingPipelineArguments):
    """Arguments related to Moses VAE model."""

    __name__ = "model_args"

    q_cell: str = field(default="gru", metadata={"help": "Encoder rnn cell type."})
    q_bidir: bool = field(
        default=True,
        metadata={"help": "Whether to add second direction in the encoder."},
    )
    q_d_h: int = field(default=256, metadata={"help": "Encoder h dimensionality."})
    q_n_layers: int = field(default=1, metadata={"help": "Encoder number of layers."})
    q_dropout: float = field(default=0.5, metadata={"help": "Encoder layers dropout."})
    d_cell: str = field(default="gru", metadata={"help": "Decoder rnn cell type."})
    d_n_layers: int = field(default=3, metadata={"help": "Decoder number of layers."})
    d_dropout: float = field(default=0, metadata={"help": "Decoder layers dropout"})
    d_z: int = field(default=128, metadata={"help": "Latent vector dimensionality"})
    d_d_h: int = field(default=512, metadata={"help": "Latent vector dimensionality"})
    freeze_embeddings: bool = field(
        default=False, metadata={"help": "If to freeze embeddings while training"}
    )


@dataclass
class MosesVAETrainingArguments(MosesTrainingArguments):
    """Arguments related to Moses VAE training."""

    n_batch: int = field(default=512, metadata={"help": "Batch size."})
    grad_clipping: int = field(
        default=50, metadata={"help": "Gradients clipping size."}
    )
    kl_start: int = field(
        default=0, metadata={"help": "Epoch to start change kl weight from."}
    )
    kl_w_start: float = field(default=0, metadata={"help": "Initial kl weight value."})
    kl_w_end: float = field(default=0.05, metadata={"help": "Maximum kl weight value."})
    lr_start: float = field(default=3 * 1e-4, metadata={"help": "Initial lr value."})
    lr_n_period: int = field(
        default=10, metadata={"help": "Epochs before first restart in SGDR."}
    )
    lr_n_restarts: int = field(
        default=10, metadata={"help": "Number of restarts in SGDR."}
    )
    lr_n_mult: int = field(
        default=1, metadata={"help": "Mult coefficient after restart in SGDR."}
    )
    lr_end: float = field(
        default=3 * 1e-4, metadata={"help": "Maximum lr weight value."}
    )
    n_last: int = field(
        default=1000, metadata={"help": "Number of iters to smooth loss calc."}
    )
    n_jobs: int = field(default=1, metadata={"help": "Number of threads."})
    n_workers: int = field(default=1, metadata={"help": "Number of workers."})



"""SMILES LSTM training pipeline from GuacaMol."""
import logging
import os
from dataclasses import dataclass, field
from typing import Any, Dict

from guacamol_baselines.smiles_lstm_hc.smiles_rnn_distribution_learner import (
    SmilesRnnDistributionLearner,
)

from ...core import TrainingPipelineArguments
from ..core import GuacaMolBaselinesTrainingPipeline

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class GuacaMolLSTMTrainingPipeline(GuacaMolBaselinesTrainingPipeline):
    """GuacaMol SMILES LSTM training pipeline."""

    def train(  # type: ignore
        self,
        training_args: Dict[str, Any],
        model_args: Dict[str, Any],
        dataset_args: Dict[str, Any],
    ) -> None:
        params = {**training_args, **model_args, **dataset_args}

        os.makedirs(params["output_dir"], exist_ok=True)

        with open(params.pop("train_smiles_filepath")) as f:
            train_list = f.readlines()

        with open(params.pop("test_smiles_filepath")) as f:
            valid_list = f.readlines()
        trainer = SmilesRnnDistributionLearner(**params)

        trainer.train(training_set=train_list, validation_set=valid_list)


@dataclass
class GuacaMolLSTMTrainingArguments(TrainingPipelineArguments):
    """Training Arguments related to SMILES LSTM trainer."""

    __name__ = "training_args"

    output_dir: str = field(metadata={"help": "Output directory."})
    batch_size: int = field(
        default=512, metadata={"help": "Size of a mini-batch for gradient descent."}
    )
    valid_every: int = field(
        default=1000, metadata={"help": "Validate every so many batches."}
    )
    n_epochs: int = field(default=10, metadata={"help": "Number of training epochs."})
    lr: float = field(default=1e-3, metadata={"help": "RNN learning rate."})


@dataclass
class GuacaMolLSTMModelArguments(TrainingPipelineArguments):
    """Arguments related to SMILES LSTM trainer."""

    __name__ = "model_args"

    hidden_size: int = field(default=512, metadata={"help": "Size of hidden layer."})
    n_layers: int = field(
        default=3, metadata={"help": "Number of layers for training."}
    )
    rnn_dropout: float = field(default=0.2, metadata={"help": "Dropout value for RNN."})
    max_len: int = field(
        default=100, metadata={"help": "Max length of a SMILES string."}
    )



"""PaccMann tests."""

from typing import ClassVar, Type

import numpy as np
import pytest

from gt4sd.algorithms.core import AlgorithmConfiguration
from gt4sd.algorithms.prediction.paccmann.core import AffinityPredictor, PaccMann
from gt4sd.algorithms.registry import ApplicationsRegistry


def get_classvar_type(class_var):
    """Extract type from ClassVar type annotation: `ClassVar[T]] -> T`."""
    return class_var.__args__[0]


@pytest.mark.parametrize(
    "config_class, algorithm_type, domain, algorithm_name",
    [
        (
            AffinityPredictor,
            "prediction",
            "materials",
            PaccMann.__name__,
        )
    ],
)
def test_config_class(
    config_class: Type[AlgorithmConfiguration],
    algorithm_type: str,
    domain: str,
    algorithm_name: str,
):
    assert config_class.algorithm_type == algorithm_type
    assert config_class.domain == domain
    assert config_class.algorithm_name == algorithm_name

    for keyword, type_annotation in config_class.__annotations__.items():
        if keyword in ("algorithm_type", "domain", "algorithm_name"):
            assert type_annotation.__origin__ is ClassVar  # type: ignore
            assert str == get_classvar_type(type_annotation)


@pytest.mark.parametrize(
    "config_class",
    [(AffinityPredictor)],
)
def test_config_instance(config_class: Type[AlgorithmConfiguration]):
    config = config_class()  # type:ignore
    assert config.algorithm_application == config_class.__name__


@pytest.mark.parametrize(
    "config, protein_targets, ligands, confidence, algorithm",
    [
        (
            AffinityPredictor,
            [
                "MVLSPADKTNVKAAWGKVGAHAGEYGAEALERMFLSFPTT",
                "MVLSPADKTNVKAAWGKVGAHAGEYGAEALERMFLSFPTT",
            ],
            [
                "CONN=COc1cc2ccccc2c11Occncc(Cl)c1N(O)O",
                "ClCCC(O1)C(C(N=C1C(=O)Nc1cccc(F)c1F)SO)C",
            ],
            True,
            PaccMann,
        )
    ],
)
def test_generation_via_import(config, protein_targets, ligands, confidence, algorithm):
    configuration = config(
        protein_targets=protein_targets, ligands=ligands, confidence=confidence
    )
    algorithm_instance = algorithm(configuration=configuration)

    predictor = configuration.get_conditional_generator(
        configuration.ensure_artifacts()
    )
    predictions, _ = predictor.predict()
    assert np.isclose(predictions[0, 0].item(), 0.5492, atol=1e-4)
    assert np.isclose(predictions[1, 0].item(), 0.4799, atol=1e-4)

    predictions = list(algorithm_instance.sample(2))
    assert np.isclose(predictions[0], 0.5492, atol=1e-4)
    assert np.isclose(predictions[1], 0.4799, atol=1e-4)


@pytest.mark.parametrize(
    "algorithm_application, protein_targets, ligands, confidence, algorithm_type, domain, algorithm_name",
    [
        (
            AffinityPredictor.__name__,
            [
                "MVLSPADKTNVKAAWGKVGAHAGEYGAEALERMFLSFPTT",
                "MVLSPADKTNVKAAWGKVGAHAGEYGAEALERMFLSFPTT",
            ],
            [
                "CONN=COc1cc2ccccc2c11Occncc(Cl)c1N(O)O",
                "ClCCC(O1)C(C(N=C1C(=O)Nc1cccc(F)c1F)SO)C",
            ],
            True,
            "prediction",
            "materials",
            PaccMann.__name__,
        ),
    ],
)
def test_generation_via_registry(
    algorithm_type,
    protein_targets,
    ligands,
    confidence,
    domain,
    algorithm_name,
    algorithm_application,
):
    algorithm = ApplicationsRegistry.get_application_instance(
        algorithm_type=algorithm_type,
        domain=domain,
        algorithm_name=algorithm_name,
        algorithm_application=algorithm_application,
        protein_targets=protein_targets,
        ligands=ligands,
        confidence=confidence,
    )
    items = list(algorithm.sample(1))
    assert len(items) == 1



"""Moses Organ training pipeline."""
import argparse
import ast
import logging
import os
from dataclasses import dataclass, field
from typing import Any, Dict

from guacamol_baselines.moses_baselines.organ_train import main
from moses.script_utils import MetricsReward

from ...core import TrainingPipelineArguments
from ..core import MosesTrainingArguments, MosesTrainingPipeline

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class MosesOrganTrainingPipeline(MosesTrainingPipeline):
    """Moses Organ training pipelines."""

    def train(  # type: ignore
        self,
        training_args: Dict[str, Any],
        model_args: Dict[str, Any],
        dataset_args: Dict[str, Any],
    ) -> None:
        """Generic training function for Moses Organ training.

        Args:
            training_args: training arguments passed to the configuration.
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.
        """
        params = {**training_args, **model_args, **dataset_args}

        os.makedirs(os.path.dirname(params["model_save"]), exist_ok=True)
        os.makedirs(os.path.dirname(params["log_file"]), exist_ok=True)
        os.makedirs(os.path.dirname(params["config_save"]), exist_ok=True)
        os.makedirs(os.path.dirname(params["vocab_save"]), exist_ok=True)
        params["addition_rewards"] = list(
            map(str.strip, params["addition_rewards"].split(","))
        )
        params["discriminator_layers"] = ast.literal_eval(
            params["discriminator_layers"]
        )

        args = argparse.Namespace(**params)
        main(args)


@dataclass
class MosesOrganTrainingArguments(MosesTrainingArguments):
    """Arguments related to Moses Organ training."""

    generator_pretrain_epochs: int = field(
        default=50, metadata={"help": "Number of epochs for generator pretraining."}
    )
    discriminator_pretrain_epochs: int = field(
        default=50, metadata={"help": "Number of epochs for discriminator pretraining."}
    )
    pg_iters: int = field(
        default=1000,
        metadata={"help": "Number of iterations for policy gradient training."},
    )
    n_batch: int = field(default=64, metadata={"help": "Size of batch."})
    lr: float = field(default=1e-4, metadata={"help": "Learning rate."})
    n_jobs: int = field(default=8, metadata={"help": "Number of threads."})
    n_workers: int = field(default=8, metadata={"help": "Number of workers."})
    clip_grad: int = field(
        default=5, metadata={"help": "Clip PG generator gradients to this value."}
    )
    rollouts: int = field(default=16, metadata={"help": "Number of rollouts."})
    generator_updates: int = field(
        default=1, metadata={"help": "Number of updates of generator per iteration."}
    )
    discriminator_updates: int = field(
        default=1,
        metadata={"help": "Number of updates of discriminator per iteration."},
    )
    discriminator_epochs: int = field(
        default=10,
        metadata={"help": "Number of epochs of discriminator per iteration."},
    )
    reward_weight: float = field(
        default=0.7, metadata={"help": "Reward weight for policy gradient training."}
    )
    addition_rewards: str = field(
        default="sa",
        metadata={
            "help": f"Comma separated list of rewards. Feasible values from: {','.join(MetricsReward.supported_metrics)}. Defaults to optimization of SA."
        },
    )
    max_length: int = field(
        default=100, metadata={"help": "Maximum length for sequence."}
    )
    n_ref_subsample: int = field(
        default=500,
        metadata={
            "help": "Number of reference molecules (sampling from training data)."
        },
    )


@dataclass
class MosesOrganModelArguments(TrainingPipelineArguments):
    """Arguments related to Moses Organ model."""

    __name__ = "model_args"

    embedding_size: int = field(
        default=32, metadata={"help": "Embedding size in generator and discriminator."}
    )
    hidden_size: int = field(
        default=512,
        metadata={"help": "Size of hidden state for lstm layers in generator."},
    )
    num_layers: int = field(
        default=2, metadata={"help": "Number of lstm layers in generator."}
    )
    dropout: float = field(
        default=0.0,
        metadata={"help": "Dropout probability for lstm layers in generator."},
    )
    discriminator_layers: str = field(
        default="[(100, 1), (200, 2), (200, 3), (200, 4), (200, 5), (100, 6), (100, 7), (100, 8), (100, 9), (100, 10), (160, 15), (160, 20)]",
        metadata={
            "help": "String representation of numbers of features for convolutional layers in discriminator."
        },
    )
    discriminator_dropout: float = field(
        default=0.0, metadata={"help": "Dropout probability for discriminator."}
    )



"""SMILES GA implementation."""

from guacamol_baselines.smiles_ga.goal_directed_generation import ChemGEGenerator


class SMILESGA:
    def __init__(
        self,
        smi_file,
        population_size: int,
        n_mutations: int,
        n_jobs: int,
        random_start: bool,
        gene_size: int,
        generations: int,
        patience: int,
    ):
        """Initialize SMILESGA.

        Args:
            smi_file: path where to load hypothesis, candidate labels and, optionally, the smiles file.
            population_size: used with n_mutations for the initial generation of smiles within the population.
            n_mutations: used with population size for the initial generation of smiles within the population.
            n_jobs: number of concurrently running jobs.
            random_start: set to True to randomly choose list of SMILES for generating optimizied molecules.
            gene_size: size of the gene which is used in creation of genes.
            generations: number of evolutionary generations.
            patience: used for early stopping if population scores remains the same after generating molecules.
        """
        self.smi_file = smi_file
        self.population_size = population_size
        self.n_mutations = n_mutations
        self.n_jobs = n_jobs
        self.random_start = random_start
        self.gene_size = gene_size
        self.generations = generations
        self.patience = patience

    def get_generator(self) -> ChemGEGenerator:
        """Create an instance of ChemGEGenerator.

        Returns:
            an instance of ChemGEGenerator.
        """
        optimiser = ChemGEGenerator(
            smi_file=self.smi_file,
            population_size=self.population_size,
            n_mutations=self.n_mutations,
            generations=self.generations,
            n_jobs=self.n_jobs,
            random_start=self.random_start,
            gene_size=self.gene_size,
            patience=self.patience,
        )
        return optimiser



"""Implementation of the KeyBERT keyword extractor."""

import os
from typing import List, Optional, Union

import torch
from keybert import KeyBERT as KeyBERTCore
from sentence_transformers import SentenceTransformer

from ....frameworks.torch import device_claim


class KeyBERT:
    """
    Keyword extractor based on [KeyBERT](https://github.com/MaartenGr/KeyBERT).
    """

    def __init__(
        self,
        resources_path: str,
        minimum_keyphrase_ngram: int,
        maximum_keyphrase_ngram: int,
        stop_words: Optional[str],
        top_n: int,
        use_maxsum: bool,
        use_mmr: bool,
        diversity: float,
        number_of_candidates: int,
        model_name: str,
        device: Optional[Union[torch.device, str]] = None,
    ):
        """Initialize KeyBERT.

        Args:
            resources_path: path where to load hypothesis, candidate labels and, optionally, the model.
            minimum_keyphrase_ngram: lower bound for phrase size.
            maximum_keyphrase_ngram: upper bound for phrase size.
            stop_words: language for the stop words removal. If not provided, no stop words removal.
            top_n: number of keywords to extract.
            use_maxsum: control usage of max sum similarity for keywords generated.
            use_mmr: control usage of max marginal relevance for keywords generated.
            diversity: diversity for the results when enabling use_mmr.
            number_of_candidates: candidates considered when enabling use_maxsum.
            model_name: name of the model to load from the cache or download from SentenceTransformers.
            device: device where the inference
                is running either as a dedicated class or a string. If not provided is inferred.
        """
        self.device = device_claim(device)
        self.resources_path = resources_path
        self.minimum_keyphrase_ngram = minimum_keyphrase_ngram
        self.maximum_keyphrase_ngram = maximum_keyphrase_ngram
        self.stop_words = stop_words
        self.top_n = top_n
        self.use_maxsum = use_maxsum
        self.use_mmr = use_mmr
        self.diversity = diversity
        self.number_of_candidates = number_of_candidates
        self.model_name = model_name
        self.load_model()

    def load_model(self) -> None:
        """Load KeyBERT model."""
        if (
            os.path.exists(self.resources_path)
            and len(os.listdir(self.resources_path)) > 0
        ):
            model_name_or_path = self.resources_path
        else:
            model_name_or_path = self.model_name
        sentence_model = SentenceTransformer(model_name_or_path, device=self.device)
        self.model = KeyBERTCore(model=sentence_model)

    def predict(self, text: str) -> List[str]:
        """Get keywords sorted by relevance.

        Args:
            text: text to extract keywords from.

        Returns:
            keywords sorted by score from highest to lowest.
        """
        return [
            keyword
            for keyword, _ in self.model.extract_keywords(
                text,
                keyphrase_ngram_range=(
                    self.minimum_keyphrase_ngram,
                    self.maximum_keyphrase_ngram,
                ),
                stop_words=self.stop_words,
                top_n=self.top_n,
                use_maxsum=self.use_maxsum,
                use_mmr=self.use_mmr,
                diversity=self.diversity,
                nr_candidates=self.number_of_candidates,
            )
        ]



"""Implementation of Regression Transformer conditional generators."""
import json
import logging
import os
from typing import Any, List, Optional, Tuple, Union

import torch
from terminator.collators import MaskedTextCollator, PropertyCollator
from terminator.inference import InferenceRT
from terminator.search import SEARCH_FACTORY, Search
from terminator.selfies import decoder
from terminator.tokenization import InferenceBertTokenizer
from transformers import AutoConfig, AutoModelWithLMHead, XLNetLMHeadModel

from ....domains.materials import Property, Sequence, validate_molecules
from ....frameworks.torch import device_claim

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class ConditionalGenerator:
    """Main interface for a regression transformer."""

    # device where the inference is running.
    device: torch.device

    # The task the RT is currently performing. Either 'regression' or 'generation'
    task: str

    # method to convert logits to predictions. Either GreedySearch or SamplingSearch
    search: Search

    # ***** Additional attributes for text generation *****
    # data collator for property prediction of self-generated items
    property_collator: PropertyCollator

    # percentage of tolerated deviation between desired and obtained property
    tolerance: float = 20

    # number of samples obtained per call
    batch_size: int = 8

    def __init__(
        self, resources_path: str, device: Optional[Union[torch.device, str]] = None
    ) -> None:
        """
        Initialize the generator.

        Args:
            resources_path: directory where to find models and parameters.
            temperature: temperature for the sampling. Defaults to 1.4.
            generated_length: maximum length of the generated molecules.
                Defaults to 100.
            samples_per_protein: number of points sampled per protein.
                It has to be greater than 1. Defaults to 10.
            device: device where the inference is running either as a dedicated class
                or a string. If not provided is inferred.
        """
        # device
        self.device = device_claim(device)

        # Set up the data preparation pipeline
        self.tokenizer = InferenceBertTokenizer.from_pretrained(
            resources_path, pad_even=False
        )
        self.collator = MaskedTextCollator(self.tokenizer)

        # Set up model: First load the pretrained XLNet model
        xlnet_model, config = self.load_model(resources_path)
        # Initialize the custom RT model
        self.model = InferenceRT(xlnet_model, self.tokenizer, config)

        # Set up inference parameters
        self.load_inference(resources_path)

    def load_model(self, resources_path: str) -> Tuple[XLNetLMHeadModel, Any]:
        """
        Loading a XLNetLMHeadModel which constitutes the base of a RT model.

        Args:
            resources_path: path to the model.

        Returns:
            XLNetLMHeadModel: base of a Regression Transformer model.
            XLNetConfig: configuration of the model.
        """
        config_name = os.path.join(resources_path, "config.json")
        config = AutoConfig.from_pretrained(config_name, mem_len=1024)
        xlnet_model = AutoModelWithLMHead.from_pretrained(
            resources_path, from_tf=False, config=config
        )
        xlnet_model.resize_token_embeddings(len(self.tokenizer))
        xlnet_model.to(self.device)
        xlnet_model.eval()
        logger.info(f"Model restored from {resources_path}")
        return xlnet_model, config

    def load_inference(self, resources_path: str) -> None:
        """
        Load and set up all parameters necessary for inference.

        Args:
            resources_path: path to the model folder.
        """
        try:
            with open(os.path.join(resources_path, "inference.json"), "r") as f:
                data = json.load(f)
            self.property = data["property_token"]
            self.property_mask_length = data["property_mask_length"][self.property]
            self.min_ = data.get("property_ranges", {}).get(self.property, [0, 1])[0]
            self.max_ = data.get("property_ranges", {}).get(self.property, [0, 1])[1]
            self.metadata = data
        except Exception:
            raise ValueError(
                f"Could not restore inference parameters from {resources_path}"
            )

    def denormalize(self, x: float, precision: int = 4) -> float:
        """
        Denormalize from [0,1] scale to original scale.

        Args:
            x: normalized value.
            precision: optional rounding precision. Defaults to 4.

        Returns:
            float: Value in regular scale.
        """
        return round(x * (self.max_ - self.min_) + self.min_, precision)

    def normalize(self, x: float, precision: int = 3) -> float:
        """
        Normalize from original scale to [0,1] scale.

        Args:
            x: unnormalized input.
            precision: optional rounding precision.

        Returns:
            float: Normalized value.
        """
        return round((x - self.min_) / (self.max_ - self.min_), precision)

    def validate_input(self, x: str) -> None:

        if self.tokenizer.expression_separator not in x:
            raise ValueError(
                f"Expression separator {self.tokenizer.expression_separator} not "
                f"found in input {x}."
            )
        if self.tokenizer.mask_token not in x:
            raise ValueError(
                f"Nothing to do, no mask to fill ({self.tokenizer.mask_token}) found"
                f"in input {x}."
            )
        if self.property not in x:
            raise ValueError(f"No property token ({self.property}) found in input")

        text_sequence = x.split(self.tokenizer.expression_separator)[-1]
        number_sequence = x[: -len(text_sequence) - 1]
        if (
            self.tokenizer.mask_token in text_sequence
            and self.tokenizer.mask_token in number_sequence
        ):
            raise ValueError(
                f"Do not mask number and text sequence at the same time like in {x}."
            )
        self.validate_input_molecule(text_sequence)

    def validate_input_molecule(self, sequence: str) -> None:
        """
        Verifies that the non-numerical part of the input is a proper sequence.

        Args:
            sequence: input sequence to be validated.
        """
        raise NotImplementedError

    def safely_determine_task(self, x: str) -> str:
        """
        Determines whether the passed sequence adheres to regression or generation task.

        Args:
            x: the user-provided input sequence for the model, inluding mask tokens.

        Raises:
            ValueError: if the sequence does not adhere to the formatting rules.

        Returns:
            str: the task, either 'regression' or 'generation'.
        """

        self.validate_input(x)
        if (
            self.tokenizer.mask_token
            in x.split(self.tokenizer.expression_separator)[-1]
        ):
            return "generation"

        if x.count(self.tokenizer.mask_token) != self.property_mask_length:
            raise ValueError(
                f"To predict {self.property} you have to mask {self.property_mask_length} times"
            )

        return "regression"

    def generate_batch_regression(self, context: Sequence) -> List[Property]:
        """
        Predict the property of a sample.

        Args:
            context: a string with a masked property, a separator and an
                entity. E.g. <stab>[MASK][MASK][MASK][MASK]|GSQEVNSGTQTYKNASPEEAERIARKAGATTWTEKGNKWEIRI.

        Returns:
            List[Property]: a list of (denormalized) predicted properties for the entity.
        """
        logger.info(f"Starting prediction for sequence {context}")

        # Prepare the batch
        inputs = self.collator([self.tokenizer(context)])
        input_ids = inputs["input_ids"].cpu()

        # Forward pass
        outputs = self.model(inputs)

        # Obtain the singular predictions
        prediction = self.search(outputs["logits"].detach())

        return self.compile_regression_result(input_ids, prediction)

    def compile_regression_result(
        self, input_ids: torch.Tensor, prediction: torch.Tensor
    ) -> List[Property]:
        """
        Postprocesses the prediction from the property task to obtain a float.

        Args:
            input_ids: 2D Tensor of shape (batch_size, sequence_length).
            prediction: 2D Tensor of shape (batch_size, sequence_length).

        Returns:
            List[Property]: list of property values.
        """
        properties = []
        for inp, pred in zip(input_ids, prediction):
            in_tokens = self.tokenizer.decode(
                inp, clean_up_tokenization_spaces=False
            ).split(" ")
            out_tokens = self.tokenizer.decode(
                pred, clean_up_tokenization_spaces=False
            ).split(" ")
            joined = self.tokenizer.get_sample_prediction(out_tokens, in_tokens)
            _, gen_prop = self.tokenizer.aggregate_tokens(joined, label_mode=False)
            properties.append(self.denormalize(gen_prop[self.property[1:-1]]))
        return properties

    def generate_batch_generation(self, sequence: Sequence) -> Tuple:
        """
        Conditionally generate sequences given a continuous property value and a fixed
        sequence. This function first conditionally generates the novel sequences and
        then predicts their properties using the RT again. Only if the predicted
        property is within the tolerance range, the novel sequence is returned.

        Args:
            sequence: the input sequence with masked tokens on the text.

        Returns:
            Tuple[Tuple[str, float]]: a tuple of tuples, each containing the generated
                sequence alongside its predicted property value.
        """

        # The property score has to be in the range [0, 1]
        sequence = self.normalize_sequence(sequence)

        logger.warning(f"Starting prediction for sequence {sequence}")

        # Prepare the batch
        tokens = self.tokenizer(sequence)
        inputs = self.collator([tokens] * self.batch_size)
        input_ids = inputs["input_ids"].clone()
        # Forward pass
        outputs = self.model(inputs)
        # Obtain model predictions via the search method
        predictions = self.search(outputs["logits"].detach()).squeeze()
        # Combine predictions with the static part to obtain the full sequences
        generations = input_ids
        generations[generations == self.tokenizer.mask_token_id] = predictions[
            generations == self.tokenizer.mask_token_id
        ]

        # Second part: Predict the properties of the just generated sequence
        _input = self.property_collator.mask_tokens(generations)
        prediction_input = {
            "input_ids": _input[0],
            "perm_mask": _input[1],
            "target_mapping": _input[2],
            "attention_mask": self.property_collator.attention_mask(generations),
        }
        # Pass through model
        property_outputs = self.model(prediction_input)
        # It's a design choice to go with greedy predictions here
        predictions = torch.argmax(property_outputs["logits"].detach(), dim=-1)
        # Obtain floating predictions
        properties = self.compile_regression_result(generations, predictions)
        # Obtain the sequences (AAS or SELFIES)
        sequences = [
            self.tokenizer.to_readable(
                "".join(
                    self.tokenizer.decode(seq, skip_special_tokens=True).split(" ")
                ).split(self.tokenizer.expression_separator)[-1]
            )
            for seq in generations
        ]
        successes: Tuple = tuple(
            filter(
                lambda x: abs(self.normalize(x[1]) - self.target_value)
                < self.tolerance,
                zip(sequences, properties),
            )
        )  # type: ignore
        logger.info(f"Successes: {successes}")
        return successes

    def normalize_sequence(self, context: Sequence) -> Sequence:
        """
        Take a sequence with a unnormalized property score and convert it to a
        sequence with a normalized score.

        Args:
            context: sequence with unnormalized property.

        Returns:
            Sequence: sequence with normalized property.
        """
        tokens = self.tokenizer.tokenize(context)
        numerical_tokens = tokens[
            tokens.index(self.property)
            + 1 : tokens.index(self.tokenizer.expression_separator)
        ]

        unnorm = self.tokenizer.floating_tokens_to_float(numerical_tokens)
        # Declard as class variable since used by other methods
        self.target_value = self.normalize(unnorm)
        norm = str(self.target_value)[: self.property_mask_length]

        tokens = (
            "".join(tokens[: tokens.index(self.property) + 1])
            + norm
            + "".join(tokens[tokens.index(self.tokenizer.expression_separator) :])
        )
        return "".join(tokens)

    @staticmethod
    def validate_numerical(sequences: List[Any]):
        """
        Validate whether a list of sequences contains only numerical values.

        Args:
            sequences: a list of hopefully only numerical values.

        Returns:
             List[Any]: a tuple containing of the validated floats and valid indexes.
        """

        items = [item if isinstance(item, float) else None for item in sequences]
        idxs = [i for i, item in enumerate(sequences) if isinstance(item, float)]
        return items, idxs


class ChemicalLanguageRT(ConditionalGenerator):
    """
    Hybrid regression and conditional molecular generation model as implemented in
    https://arxiv.org/abs/2202.01338. It generates molecules with a desired solubility
    (ESOL) score or predicts the ESOL of a given molecule.
    For details on the ESOL task see: https://pubs.acs.org/doi/10.1021/ci034243x

    Attributes:
        resources_path: path to the model.
        context: user-specified input text for the model.
        search: search key to instantiate a search via terminator.search.SEARCH_FACTORY.
        temperature: the temperature parameter in case of a `sample` search.
        batch_size: the batch size for the model, applicable only to generative task.
        tolerance: the tolerance for the property of the generated molecules.
    """

    def __init__(
        self,
        resources_path: str,
        context: str,
        search: str = "sample",
        temperature: float = 1.4,
        batch_size: int = 8,
        tolerance: float = 20.0,
        device: Optional[Union[torch.device, str]] = None,
    ) -> None:
        """
        Initialize the molecule generator.

        Args:
            resources_path: directory where to find models and parameters.
            search: search key to instantiate a search, defaults to `sample`.
            temperature: temperature for the sampling. Defaults to 1.4.
            batch_size: number of points sampled per call. Defaults to 8.
            tolerance: the tolerance for the property of the generated molecules.
                Given in percent. Defaults to 20.
            device: device where the inference s running either as a dedicated class
                or a string. If not provided is inferred.
        """
        super().__init__(device=device, resources_path=resources_path)

        # Validate input and determine task
        self.task = self.safely_determine_task(context)

        # Console outputs for usage of search methods
        if search == "sample" and self.task == "regression":
            logger.warning("For regression task, greedy search is recommended")
        elif search == "greedy" and self.task == "generation":
            logger.warning("For generation task, sample search is recommended")

        if search not in SEARCH_FACTORY.keys():
            raise KeyError(f"Pick a search of {SEARCH_FACTORY.keys()} not: {search}.")
        self.search = SEARCH_FACTORY[search](temperature=temperature)

        # Based on task, set the correct generate_batch method
        if self.task == "regression":
            self.generate_batch = self.generate_batch_regression
        elif self.task == "generation":
            self.generate_batch = self.generate_batch_generation  # type: ignore
            self.batch_size = batch_size
            self.property_collator = PropertyCollator(
                tokenizer=self.tokenizer,
                property_tokens=[self.property],
                num_tokens_to_mask=[-1],
                ignore_errors=False,
            )
            # Tolerance on [0,1] scale
            self.tolerance = tolerance / 100.0
        else:
            raise ValueError(f"Unknown task: {self.task}")

    def validate_input_molecule(self, sequence: str) -> None:
        """
        Verifies that the non-numerical part of the input sequence is a SELFIES.

        Args:
            sequence: input sequence to be validated.
        """
        # Fractional molecules based on non-masked parts of the SELFIES sequence
        smis = list(map(decoder, sequence.split(self.tokenizer.mask_token)))
        if -1 in smis:
            raise ValueError(f"Invalid sequence: {sequence}")

    def validate_output(self, sequences: List[Any]) -> Tuple[List[Any], List[int]]:
        """
        Validate the output of the RT model.

        Args:
            sequences: list of sequences to be validated.

        Returns:
            A tuple of validated items (Chem.rdchem.Mol in the case of a generation task
                floating values otherwise) and a list of valid indexes.
        """

        if self.task == "regression":
            return self.validate_numerical(sequences)
        else:
            # Convert SELFIES to SMILES
            smiles_list = list(
                filter(lambda x: x is not None, list(zip(*sequences))[0])
            )
            if smiles_list == []:
                return ([None], [-1])
            return validate_molecules(smiles_list=smiles_list)  # type: ignore


class ProteinLanguageRT(ConditionalGenerator):
    """
    Hybrid regression and conditional protein generation model as implemented in
    https://arxiv.org/abs/2202.01338. It generates peptides with a desired stability
    score or predicts the stability score of a given molecule.
    For details on the stability task see: https://doi.org/10.1126/science.aan0693

    Attributes:
        resources_path: path to the model.
        context: user-specified input text for the model.
        search: search key to instantiate a search via terminator.search.SEARCH_FACTORY.
        temperature: the temperature parameter in case of a `sample` search.
        batch_size: the batch size for the model, applicable only to generative task.
        tolerance: the tolerance for the property of the generated molecules.
    """

    def __init__(
        self,
        resources_path: str,
        context: str,
        search: str = "sample",
        temperature: float = 1.4,
        batch_size: int = 32,
        tolerance: float = 20.0,
        device: Optional[Union[torch.device, str]] = None,
    ) -> None:
        """
        Initialize the protein generator.

        Args:
            resources_path: directory where to find models and parameters.
            search: search key to instantiate a search, defaults to `sample`.
            temperature: temperature for the sampling. Defaults to 1.4.
            batch_size: number of points sampled per call. Defaults to 8.
            tolerance: the tolerance for the property of the generated molecules.
                Given in percent. Defaults to 20.
            device: device where the inference s running either as a dedicated class
                or a string. If not provided is inferred.
        """
        super().__init__(device=device, resources_path=resources_path)

        # Validate input and determine task
        self.task = self.safely_determine_task(context)

        # Console outputs for usage of search methods
        if search == "sample" and self.task == "regression":
            logger.warning("For regression task, greedy search is recommended")
        elif search == "greedy" and self.task == "generation":
            logger.warning("For generation task, sample search is recommended")
        if search not in SEARCH_FACTORY.keys():
            raise KeyError(f"Pick a search of {SEARCH_FACTORY.keys()} not: {search}.")
        self.search = SEARCH_FACTORY[search](temperature=temperature)

        # Based on task, set the correct generate_batch method
        if self.task == "regression":
            self.generate_batch = self.generate_batch_regression
        elif self.task == "generation":
            self.generate_batch = self.generate_batch_generation  # type: ignore
            self.batch_size = batch_size
            self.property_collator = PropertyCollator(
                tokenizer=self.tokenizer,
                property_tokens=[self.property],
                num_tokens_to_mask=[-1],
                ignore_errors=False,
            )
            # Tolerance on [0,1] scale
            self.tolerance = tolerance / 100.0
        else:
            raise ValueError(f"Unknown task: {self.task}")

    def validate_input_molecule(self, sequence: str) -> None:
        """
        Verifies that the non-numerical part of the input sequence is a valid AAS.

        Args:
            sequence: input sequence to be validated.
        """
        if sequence != sequence.upper():
            raise ValueError(
                f"Sequence {sequence} does not follow IUPAC convention for AAS"
            )

    def validate_output(self, sequences: Any) -> Tuple[List[Any], List[int]]:
        """
        Validate the output of the RT model.

        Args:
            sequences: list of sequences to be validated.

        Returns:
            A tuple of validated items and a list of valid indexes.
        """

        if self.task == "regression":
            return self.validate_numerical(sequences)
        else:
            items = [
                item
                if (
                    (
                        item[0] == item[0].upper()
                        and self.tokenizer.mask_token not in item[0]
                        and not any([s.isdigit() for s in item[0]])
                    )
                    and isinstance(item[1], float)
                )
                else None
                for item in sequences
            ]
            idxs = [i for i, item in enumerate(sequences) if item in items]
            return items, idxs



"""Model initialization module."""

from typing import Dict, Type

from .activation import ACTIVATION_FACTORY  # noqa: F401
from .base_model import GranularBaseModel, GranularEncoderDecoderModel
from .loss import LOSS_FACTORY  # noqa: F401
from .mlp_auto_encoder import MlpAutoEncoder
from .mlp_predictor import MlpPredictor
from .no_encoding import NoEncoding
from .vae_mlp import VaeMlp
from .vae_rnn import VaeRnn
from .vae_trans import VaeTrans

ARCHITECTURE_FACTORY: Dict[str, Type[GranularBaseModel]] = {
    "vae_rnn": VaeRnn,
    "vae_trans": VaeTrans,
    "mlp_predictor": MlpPredictor,
    "no_encoding": NoEncoding,
    "mlp_autoencoder": MlpAutoEncoder,
    "vae_mlp": VaeMlp,
}
AUTOENCODER_ARCHITECTURES = set(
    [
        model_type
        for model_type, model_class in ARCHITECTURE_FACTORY.items()
        if issubclass(model_class, GranularEncoderDecoderModel)
    ]
)



"""Implementation details for PaccMann vanilla generator trained on polymer building blocks (catalysts/monomers)."""

import json
import os
from typing import List, Optional, Union

import torch
from rdkit import Chem, RDLogger
from paccmann_chemistry.models.vae import StackGRUDecoder, StackGRUEncoder, TeacherVAE
from paccmann_chemistry.utils import get_device
from paccmann_chemistry.utils.search import SamplingSearch
from pytoda.smiles.smiles_language import SMILESLanguage
from pytoda.smiles.transforms import Selfies, SMILESToTokenIndexes
from pytoda.transforms import Compose, ToTensor

from ....frameworks.torch import device_claim

RDLogger.DisableLog("rdApp.*")


class Generator:
    def __init__(
        self,
        resources_path: str,
        generated_length: int = 100,
        batch_size: int = 32,
        device: Optional[Union[torch.device, str]] = None,
    ):
        """Initialize the encoder/decoder generative model.

        Args:
            resources_path: directory where to find models and parameters.
            generated_length: length of the generated molecule in tokens. Defaults to 100.
            batch_size: size of the batch. Defaults to 1.
            device: device where the inference is running either as a dedicated class or a string.
                If not provided is inferred.
        """
        self.device = device_claim(device)
        self.generated_length = generated_length
        self.batch_size = batch_size
        self.resources_path = resources_path
        self.load_pretrained_paccmann(
            os.path.join(self.resources_path, "params.json"),
            os.path.join(self.resources_path, "smiles_language.pkl"),
            os.path.join(self.resources_path, "weights.pt"),
            self.batch_size,
        )

    def load_pretrained_paccmann(
        self, params_file: str, lang_file: str, weights_file: str, batch_size: int
    ) -> None:
        """Load a pretrained PaccMann model.

        Args:
            params_file: file for the parameters.
            lang_file: language file.
            weights_file: serialized weights file.
            batch_size: size of the batch.
        """
        params = dict()
        with open(params_file, "r") as f:
            params.update(json.load(f))
        params["batch_mode"] = "Padded"
        params["batch_size"] = batch_size

        self.selfies = params.get("selfies", False)

        self.device = get_device()
        self.smiles_language = SMILESLanguage.load(lang_file)

        self.gru_encoder = StackGRUEncoder(params).to(self.device)
        self.gru_decoder = StackGRUDecoder(params).to(self.device)
        self.gru_vae = TeacherVAE(self.gru_encoder, self.gru_decoder).to(self.device)
        self.gru_vae.load_state_dict(torch.load(weights_file, map_location=self.device))
        self.gru_vae.eval()

        transforms = []
        if self.selfies:
            transforms += [Selfies()]
        transforms += [SMILESToTokenIndexes(smiles_language=self.smiles_language)]
        transforms += [ToTensor(device=self.device)]
        self.transform = Compose(transforms)

    def decode(
        self, latent_z: torch.Tensor, search: SamplingSearch = SamplingSearch()
    ) -> List[int]:
        """Decodes a sequence of tokens given a position in the latent space.

        Args:
            latent_z: a batch size x latent size tensor.
            search: defaults to sampling multinomial search.

        Returns:
            list of list of token indices.
        """
        latent_z = latent_z.view(1, latent_z.shape[0], latent_z.shape[1]).float()
        molecule_iter = self.gru_vae.generate(
            latent_z,
            prime_input=torch.tensor([self.smiles_language.start_index]).to(
                self.device
            ),
            end_token=torch.tensor([self.smiles_language.stop_index]).to(self.device),
            generate_len=self.generated_length,
            search=search,
        )
        return [
            [self.smiles_language.start_index] + m.cpu().detach().tolist()
            for m in molecule_iter
        ]

    def sample(self) -> List[str]:
        """Sample random molecules.

        Returns:
            sampled molecule (SMILES).
        """
        mol: List[str] = []
        while len(mol) < 1:
            indexes = self.decode(
                torch.randn(
                    self.batch_size, self.gru_decoder.latent_dim, device=self.device
                )
            )
            mol = [self.smiles_language.token_indexes_to_smiles(m) for m in indexes]
            mol = [m for m in mol if Chem.MolFromSmiles(m) is not None and m != ""]
        return mol
